@@ -1,29 +1,21 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2017 Intel Corporation
 
-    This file is part of Threading Building Blocks.
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
+
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 
-    Threading Building Blocks is free software; you can redistribute it
-    and/or modify it under the terms of the GNU General Public License
-    version 2 as published by the Free Software Foundation.
 
-    Threading Building Blocks is distributed in the hope that it will be
-    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with Threading Building Blocks; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    As a special exception, you may use this file as part of a free software
-    library without restriction.  Specifically, if other files instantiate
-    templates or use macros or inline functions from this file, or you compile
-    this file and link it with other files to produce an executable, this
-    file does not by itself cause the resulting executable to be covered by
-    the GNU General Public License.  This exception does not however
-    invalidate any other reasons why the executable file might be covered by
-    the GNU General Public License.
 */
 
 #ifndef __TBB__concurrent_queue_impl_H
@@ -41,19 +33,9 @@
 #include "../tbb_exception.h"
 #include "../tbb_profiling.h"
 #include <new>
-
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    // Suppress "C++ exception handler used, but unwind semantics are not enabled" warning in STL headers
-    #pragma warning (push)
-    #pragma warning (disable: 4530)
-#endif
-
+#include __TBB_STD_SWAP_HEADER
 #include <iterator>
 
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    #pragma warning (pop)
-#endif
-
 namespace tbb {
 
 #if !__TBB_TEMPLATE_FRIENDS_BROKEN
@@ -65,9 +47,6 @@ template<typename T, typename A> class concurrent_queue;
 
 template<typename T, typename A> class concurrent_bounded_queue;
 
-namespace deprecated {
-template<typename T, typename A> class concurrent_queue;
-}
 #endif
 
 //! For internal use only.
@@ -83,6 +62,7 @@ typedef size_t ticket;
 template<typename T> class micro_queue ;
 template<typename T> class micro_queue_pop_finalizer ;
 template<typename T> class concurrent_queue_base_v3;
+template<typename T> struct concurrent_queue_rep;
 
 //! parts of concurrent_queue_rep that do not have references to micro_queue
 /**
@@ -153,6 +133,9 @@ private:
     The caller is expected to zero-initialize it. */
 template<typename T>
 class micro_queue : no_copy {
+public:
+    typedef void (*item_constructor_t)(T* location, const void* src);
+private:
     typedef concurrent_queue_rep_base::page page;
 
     //! Class used to ensure exception-safety of method "pop"
@@ -163,18 +146,21 @@ class micro_queue : no_copy {
         ~destroyer() {my_value.~T();}
     };
 
-    void copy_item( page& dst, size_t index, const void* src ) {
-        new( &get_ref(dst,index) ) T(*static_cast<const T*>(src));
+    void copy_item( page& dst, size_t dindex, const void* src, item_constructor_t construct_item ) {
+        construct_item( &get_ref(dst, dindex), src );
     }
 
-    void copy_item( page& dst, size_t dindex, const page& src, size_t sindex ) {
-        new( &get_ref(dst,dindex) ) T( get_ref(const_cast<page&>(src),sindex) );
+    void copy_item( page& dst, size_t dindex, const page& src, size_t sindex,
+        item_constructor_t construct_item )
+    {
+        T& src_item = get_ref( const_cast<page&>(src), sindex );
+        construct_item( &get_ref(dst, dindex), static_cast<const void*>(&src_item) );
     }
 
     void assign_and_destroy_item( void* dst, page& src, size_t index ) {
         T& from = get_ref(src,index);
         destroyer d(from);
-        *static_cast<T*>(dst) = from;
+        *static_cast<T*>(dst) = tbb::internal::move( from );
     }
 
     void spin_wait_until_my_turn( atomic<ticket>& counter, ticket k, concurrent_queue_rep_base& rb ) const ;
@@ -203,13 +189,16 @@ public:
 
     spin_mutex page_mutex;
 
-    void push( const void* item, ticket k, concurrent_queue_base_v3<T>& base ) ;
+    void push( const void* item, ticket k, concurrent_queue_base_v3<T>& base,
+        item_constructor_t construct_item ) ;
 
     bool pop( void* dst, ticket k, concurrent_queue_base_v3<T>& base ) ;
 
-    micro_queue& assign( const micro_queue& src, concurrent_queue_base_v3<T>& base ) ;
+    micro_queue& assign( const micro_queue& src, concurrent_queue_base_v3<T>& base,
+        item_constructor_t construct_item ) ;
 
-    page* make_copy( concurrent_queue_base_v3<T>& base, const page* src_page, size_t begin_in_page, size_t end_in_page, ticket& g_index ) ;
+    page* make_copy( concurrent_queue_base_v3<T>& base, const page* src_page, size_t begin_in_page,
+        size_t end_in_page, ticket& g_index, item_constructor_t construct_item ) ;
 
     void invalidate_page_and_rethrow( ticket k ) ;
 };
@@ -227,7 +216,9 @@ void micro_queue<T>::spin_wait_until_my_turn( atomic<ticket>& counter, ticket k,
 }
 
 template<typename T>
-void micro_queue<T>::push( const void* item, ticket k, concurrent_queue_base_v3<T>& base ) {
+void micro_queue<T>::push( const void* item, ticket k, concurrent_queue_base_v3<T>& base,
+    item_constructor_t construct_item )
+{
     k &= -concurrent_queue_rep_base::n_queue;
     page* p = NULL;
     size_t index = modulo_power_of_two( k/concurrent_queue_rep_base::n_queue, base.my_rep->items_per_page);
@@ -243,7 +234,7 @@ void micro_queue<T>::push( const void* item, ticket k, concurrent_queue_base_v3<
         p->next = NULL;
     }
 
-    if( tail_counter!=k ) spin_wait_until_my_turn( tail_counter, k, *base.my_rep );
+    if( tail_counter != k ) spin_wait_until_my_turn( tail_counter, k, *base.my_rep );
     call_itt_notify(acquired, &tail_counter);
 
     if( p ) {
@@ -257,8 +248,9 @@ void micro_queue<T>::push( const void* item, ticket k, concurrent_queue_base_v3<
     } else {
         p = tail_page;
     }
+
     __TBB_TRY {
-        copy_item( *p, index, item );
+        copy_item( *p, index, item, construct_item );
         // If no exception was thrown, mark item as present.
         itt_hide_store_word(p->mask,  p->mask | uintptr_t(1)<<index);
         call_itt_notify(releasing, &tail_counter);
@@ -278,15 +270,15 @@ bool micro_queue<T>::pop( void* dst, ticket k, concurrent_queue_base_v3<T>& base
     call_itt_notify(acquired, &head_counter);
     if( tail_counter==k ) spin_wait_while_eq( tail_counter, k );
     call_itt_notify(acquired, &tail_counter);
-    page& p = *head_page;
-    __TBB_ASSERT( &p, NULL );
+    page *p = head_page;
+    __TBB_ASSERT( p, NULL );
     size_t index = modulo_power_of_two( k/concurrent_queue_rep_base::n_queue, base.my_rep->items_per_page );
     bool success = false;
     {
-        micro_queue_pop_finalizer<T> finalizer( *this, base, k+concurrent_queue_rep_base::n_queue, index==base.my_rep->items_per_page-1 ? &p : NULL );
-        if( p.mask & uintptr_t(1)<<index ) {
+        micro_queue_pop_finalizer<T> finalizer( *this, base, k+concurrent_queue_rep_base::n_queue, index==base.my_rep->items_per_page-1 ? p : NULL );
+        if( p->mask & uintptr_t(1)<<index ) {
             success = true;
-            assign_and_destroy_item( dst, p, index );
+            assign_and_destroy_item( dst, *p, index );
         } else {
             --base.my_rep->n_invalid_entries;
         }
@@ -295,10 +287,11 @@ bool micro_queue<T>::pop( void* dst, ticket k, concurrent_queue_base_v3<T>& base
 }
 
 template<typename T>
-micro_queue<T>& micro_queue<T>::assign( const micro_queue<T>& src, concurrent_queue_base_v3<T>& base ) {
+micro_queue<T>& micro_queue<T>::assign( const micro_queue<T>& src, concurrent_queue_base_v3<T>& base,
+    item_constructor_t construct_item )
+{
     head_counter = src.head_counter;
     tail_counter = src.tail_counter;
-    page_mutex   = src.page_mutex;
 
     const page* srcp = src.head_page;
     if( is_valid_page(srcp) ) {
@@ -308,12 +301,12 @@ micro_queue<T>& micro_queue<T>::assign( const micro_queue<T>& src, concurrent_qu
             size_t index = modulo_power_of_two( head_counter/concurrent_queue_rep_base::n_queue, base.my_rep->items_per_page );
             size_t end_in_first_page = (index+n_items<base.my_rep->items_per_page)?(index+n_items):base.my_rep->items_per_page;
 
-            head_page = make_copy( base, srcp, index, end_in_first_page, g_index );
+            head_page = make_copy( base, srcp, index, end_in_first_page, g_index, construct_item );
             page* cur_page = head_page;
 
             if( srcp != src.tail_page ) {
                 for( srcp = srcp->next; srcp!=src.tail_page; srcp=srcp->next ) {
-                    cur_page->next = make_copy( base, srcp, 0, base.my_rep->items_per_page, g_index );
+                    cur_page->next = make_copy( base, srcp, 0, base.my_rep->items_per_page, g_index, construct_item );
                     cur_page = cur_page->next;
                 }
 
@@ -321,7 +314,7 @@ micro_queue<T>& micro_queue<T>::assign( const micro_queue<T>& src, concurrent_qu
                 size_t last_index = modulo_power_of_two( tail_counter/concurrent_queue_rep_base::n_queue, base.my_rep->items_per_page );
                 if( last_index==0 ) last_index = base.my_rep->items_per_page;
 
-                cur_page->next = make_copy( base, srcp, 0, last_index, g_index );
+                cur_page->next = make_copy( base, srcp, 0, last_index, g_index, construct_item );
                 cur_page = cur_page->next;
             }
             tail_page = cur_page;
@@ -352,14 +345,17 @@ void micro_queue<T>::invalidate_page_and_rethrow( ticket k ) {
 }
 
 template<typename T>
-concurrent_queue_rep_base::page* micro_queue<T>::make_copy( concurrent_queue_base_v3<T>& base, const concurrent_queue_rep_base::page* src_page, size_t begin_in_page, size_t end_in_page, ticket& g_index ) {
+concurrent_queue_rep_base::page* micro_queue<T>::make_copy( concurrent_queue_base_v3<T>& base,
+    const concurrent_queue_rep_base::page* src_page, size_t begin_in_page, size_t end_in_page,
+    ticket& g_index, item_constructor_t construct_item )
+{
     concurrent_queue_page_allocator& pa = base;
     page* new_page = pa.allocate_page();
     new_page->next = NULL;
     new_page->mask = src_page->mask;
     for( ; begin_in_page!=end_in_page; ++begin_in_page, ++g_index )
         if( new_page->mask & uintptr_t(1)<<begin_in_page )
-            copy_item( *new_page, begin_in_page, *src_page, begin_in_page );
+            copy_item( *new_page, begin_in_page, *src_page, begin_in_page, construct_item );
     return new_page;
 }
 
@@ -427,6 +423,7 @@ struct concurrent_queue_rep : public concurrent_queue_rep_base {
  */
 template<typename T>
 class concurrent_queue_base_v3: public concurrent_queue_page_allocator {
+private:
     //! Internal representation
     concurrent_queue_rep<T>* my_rep;
 
@@ -440,14 +437,15 @@ protected:
 
 private:
     typedef typename micro_queue<T>::padded_page padded_page;
+    typedef typename micro_queue<T>::item_constructor_t item_constructor_t;
 
-    /* override */ virtual page *allocate_page() {
+    virtual page *allocate_page() __TBB_override {
         concurrent_queue_rep<T>& r = *my_rep;
         size_t n = sizeof(padded_page) + (r.items_per_page-1)*sizeof(T);
         return reinterpret_cast<page*>(allocate_block ( n ));
     }
 
-    /* override */ virtual void deallocate_page( concurrent_queue_rep_base::page *p ) {
+    virtual void deallocate_page( concurrent_queue_rep_base::page *p ) __TBB_override {
         concurrent_queue_rep<T>& r = *my_rep;
         size_t n = sizeof(padded_page) + (r.items_per_page-1)*sizeof(T);
         deallocate_block( reinterpret_cast<void*>(p), n );
@@ -462,7 +460,7 @@ private:
 protected:
     concurrent_queue_base_v3();
 
-    /* override */ virtual ~concurrent_queue_base_v3() {
+    virtual ~concurrent_queue_base_v3() {
 #if TBB_USE_ASSERT
         size_t nq = my_rep->n_queue;
         for( size_t i=0; i<nq; i++ )
@@ -472,10 +470,10 @@ protected:
     }
 
     //! Enqueue item at tail of queue
-    void internal_push( const void* src ) {
-        concurrent_queue_rep<T>& r = *my_rep;
-        ticket k = r.tail_counter++;
-        r.choose(k).push( src, k, *this );
+    void internal_push( const void* src, item_constructor_t construct_item ) {
+         concurrent_queue_rep<T>& r = *my_rep;
+         ticket k = r.tail_counter++;
+         r.choose(k).push( src, k, *this, construct_item );
     }
 
     //! Attempt to dequeue item from queue.
@@ -497,8 +495,15 @@ protected:
         throw_exception( eid_bad_alloc );
     }
 
-    //! copy internal representation
-    void assign( const concurrent_queue_base_v3& src ) ;
+    //! copy or move internal representation
+    void assign( const concurrent_queue_base_v3& src, item_constructor_t construct_item ) ;
+
+#if __TBB_CPP11_RVALUE_REF_PRESENT
+    //! swap internal representation
+    void internal_swap( concurrent_queue_base_v3& src ) {
+        std::swap( my_rep, src.my_rep );
+    }
+#endif /* __TBB_CPP11_RVALUE_REF_PRESENT */
 };
 
 template<typename T>
@@ -585,18 +590,20 @@ void concurrent_queue_base_v3<T>::internal_finish_clear() {
 }
 
 template<typename T>
-void concurrent_queue_base_v3<T>::assign( const concurrent_queue_base_v3& src ) {
+void concurrent_queue_base_v3<T>::assign( const concurrent_queue_base_v3& src,
+    item_constructor_t construct_item )
+{
     concurrent_queue_rep<T>& r = *my_rep;
     r.items_per_page = src.my_rep->items_per_page;
 
-    // copy concurrent_queue_rep.
+    // copy concurrent_queue_rep data
     r.head_counter = src.my_rep->head_counter;
     r.tail_counter = src.my_rep->tail_counter;
     r.n_invalid_entries = src.my_rep->n_invalid_entries;
 
-    // copy micro_queues
-    for( size_t i = 0; i<r.n_queue; ++i )
-        r.array[i].assign( src.my_rep->array[i], *this);
+    // copy or move micro_queues
+    for( size_t i = 0; i < r.n_queue; ++i )
+        r.array[i].assign( src.my_rep->array[i], *this, construct_item);
 
     __TBB_ASSERT( r.head_counter==src.my_rep->head_counter && r.tail_counter==src.my_rep->tail_counter,
             "the source concurrent queue should not be concurrently modified." );
@@ -743,10 +750,10 @@ class concurrent_queue_iterator: public concurrent_queue_iterator_base_v3<typena
     template<typename T, class A>
     friend class ::tbb::strict_ppl::concurrent_queue;
 #else
-public: // workaround for MSVC
+public:
 #endif
     //! Construct iterator pointing to head of queue.
-    concurrent_queue_iterator( const concurrent_queue_base_v3<Value>& queue ) :
+    explicit concurrent_queue_iterator( const concurrent_queue_base_v3<typename tbb_remove_cv<Value>::type>& queue ) :
         concurrent_queue_iterator_base_v3<typename tbb_remove_cv<Value>::type>(queue)
     {
     }
@@ -754,6 +761,8 @@ public: // workaround for MSVC
 public:
     concurrent_queue_iterator() {}
 
+    /** If Value==Container::value_type, then this routine is the copy constructor.
+        If Value==const Container::value_type, then this routine is a conversion constructor. */
     concurrent_queue_iterator( const concurrent_queue_iterator<Container,typename Container::value_type>& other ) :
         concurrent_queue_iterator_base_v3<typename tbb_remove_cv<Value>::type>(other)
     {}
@@ -814,6 +823,7 @@ template<typename Container, typename Value> class concurrent_queue_iterator;
 /** Type-independent portion of concurrent_queue.
     @ingroup containers */
 class concurrent_queue_base_v3: no_copy {
+private:
     //! Internal representation
     concurrent_queue_rep* my_rep;
 
@@ -838,6 +848,8 @@ protected:
     //! Size of an item
     size_t item_size;
 
+    enum copy_specifics { copy, move };
+
 #if __TBB_PROTECTED_NESTED_CLASS_BROKEN
 public:
 #endif
@@ -858,7 +870,7 @@ protected:
     __TBB_EXPORTED_METHOD concurrent_queue_base_v3( size_t item_size );
     virtual __TBB_EXPORTED_METHOD ~concurrent_queue_base_v3();
 
-    //! Enqueue item at tail of queue
+    //! Enqueue item at tail of queue using copy operation
     void __TBB_EXPORTED_METHOD internal_push( const void* src );
 
     //! Dequeue item from head of queue
@@ -867,7 +879,7 @@ protected:
     //! Abort all pending queue operations
     void __TBB_EXPORTED_METHOD internal_abort();
 
-    //! Attempt to enqueue item onto queue.
+    //! Attempt to enqueue item onto queue using copy operation
     bool __TBB_EXPORTED_METHOD internal_push_if_not_full( const void* src );
 
     //! Attempt to dequeue item from queue.
@@ -899,10 +911,49 @@ protected:
     //! copy internal representation
     void __TBB_EXPORTED_METHOD assign( const concurrent_queue_base_v3& src ) ;
 
+#if __TBB_CPP11_RVALUE_REF_PRESENT
+    //! swap queues
+    void internal_swap( concurrent_queue_base_v3& src ) {
+        std::swap( my_capacity, src.my_capacity );
+        std::swap( items_per_page, src.items_per_page );
+        std::swap( item_size, src.item_size );
+        std::swap( my_rep, src.my_rep );
+    }
+#endif /* __TBB_CPP11_RVALUE_REF_PRESENT */
+
+    //! Enqueues item at tail of queue using specified operation (copy or move)
+    void internal_insert_item( const void* src, copy_specifics op_type );
+
+    //! Attempts to enqueue at tail of queue using specified operation (copy or move)
+    bool internal_insert_if_not_full( const void* src, copy_specifics op_type );
+
+    //! Assigns one queue to another using specified operation (copy or move)
+    void internal_assign( const concurrent_queue_base_v3& src, copy_specifics op_type );
 private:
     virtual void copy_page_item( page& dst, size_t dindex, const page& src, size_t sindex ) = 0;
 };
 
+//! For internal use only.
+/** Backward compatible modification of concurrent_queue_base_v3
+    @ingroup containers */
+class concurrent_queue_base_v8: public concurrent_queue_base_v3 {
+protected:
+    concurrent_queue_base_v8( size_t item_sz ) : concurrent_queue_base_v3( item_sz ) {}
+
+    //! move items
+    void __TBB_EXPORTED_METHOD move_content( concurrent_queue_base_v8& src ) ;
+
+    //! Attempt to enqueue item onto queue using move operation
+    bool __TBB_EXPORTED_METHOD internal_push_move_if_not_full( const void* src );
+
+    //! Enqueue item at tail of queue using move operation
+    void __TBB_EXPORTED_METHOD internal_push_move( const void* src );
+private:
+    friend struct micro_queue;
+    virtual void move_page_item( page& dst, size_t dindex, const page& src, size_t sindex ) = 0;
+    virtual void move_item( page& dst, size_t index, const void* src ) = 0;
+};
+
 //! Type-independent portion of concurrent_queue_iterator.
 /** @ingroup containers */
 class concurrent_queue_iterator_base_v3 {
@@ -955,17 +1006,15 @@ template<typename Container, typename Value>
 class concurrent_queue_iterator: public concurrent_queue_iterator_base,
         public std::iterator<std::forward_iterator_tag,Value> {
 
-#if !defined(_MSC_VER) || defined(__INTEL_COMPILER)
+#if !__TBB_TEMPLATE_FRIENDS_BROKEN
     template<typename T, class A>
     friend class ::tbb::concurrent_bounded_queue;
-
-    template<typename T, class A>
-    friend class ::tbb::deprecated::concurrent_queue;
 #else
-public: // workaround for MSVC
+public:
 #endif
+
     //! Construct iterator pointing to head of queue.
-    concurrent_queue_iterator( const concurrent_queue_base_v3& queue ) :
+    explicit concurrent_queue_iterator( const concurrent_queue_base_v3& queue ) :
         concurrent_queue_iterator_base_v3(queue,__TBB_offsetof(concurrent_queue_base_v3::padded_page<Value>,last))
     {
     }
