@@ -1,21 +1,21 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2017 Intel Corporation
+
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
+
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
+
+
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
 */
 
 //#define DO_SCALABLEALLOC
@@ -93,9 +93,9 @@ public:
             dv[len-i-1] = timer0.diff_time(tv[i]);
             tv[i].mark_time();
         }
-        stat.AddStatisticValue( key[2], "1total, ms", "%.3f", timer2.get_time()*1000.0 );
-        stat.AddStatisticValue( key[1], "1total, ms", "%.3f", timer1.diff_time(timer2)*1000.0 );
-        stat.AddStatisticValue( key[0], "1total, ms", "%.3f", timer0.diff_time(timer1)*1000.0 );
+        stat.AddStatisticValue( key[2], "1total, ms", "%.3f", timer2.get_time()*1e+3 );
+        stat.AddStatisticValue( key[1], "1total, ms", "%.3f", timer1.diff_time(timer2)*1e+3 );
+        stat.AddStatisticValue( key[0], "1total, ms", "%.3f", timer0.diff_time(timer1)*1e+3 );
         //allocator statistics
         stat.AddStatisticValue( key[0], "2total allocations", "%d", int(timers_vector_t::allocator_type::allocations) );
         stat.AddStatisticValue( key[1], "2total allocations", "%d", int(values_vector_t::allocator_type::allocations) );
@@ -104,13 +104,13 @@ public:
         stat.AddStatisticValue( key[1], "3total alloc#items", "%d", int(values_vector_t::allocator_type::items_allocated) );
         stat.AddStatisticValue( key[2], "3total alloc#items", "%d",  0);
         //remarks
-        stat.AddStatisticValue( key[0], "9note", "segment creation time, ns:");
-        stat.AddStatisticValue( key[2], "9note", "average op-time per item, ns:");
+        stat.AddStatisticValue( key[0], "9note", "segment creation time, us:");
+        stat.AddStatisticValue( key[2], "9note", "average op-time per item, us:");
         Timer last_timer(timer2); double last_value = 0;
         for (size_t j = 0, i = 2; i < len; i *= 2, j++) {
-            stat.AddRoundResult( key[0], (dv[len-i-1]-last_value)*1000000.0 );
+            stat.AddRoundResult( key[0], (dv[len-i-1]-last_value)*1e+6 );
             last_value = dv[len-i-1];
-            stat.AddRoundResult( key[2], last_timer.diff_time(tv[i])/double(i)*1000000.0 );
+            stat.AddRoundResult( key[2], last_timer.diff_time(tv[i])/double(i)*1e+6 );
             last_timer = tv[i];
             stat.SetRoundTitle(j, i);
         }
@@ -154,10 +154,10 @@ public:
             std::sort(v2.rbegin(), v2.rend());
             std::set_intersection(v1.begin(), v1.end(), v2.rbegin(), v2.rend(), v1.begin());
             Timer timer3;
-            stat.AddRoundResult( proc_key, timer2.diff_time(timer3)*1000.0 );
-            stat.AddRoundResult( fill_key, timer1.diff_time(timer2)*1000.0 );
-            stat.AddRoundResult( init_key, timer0.diff_time(timer1)*1000.0 );
-            stat.AddRoundResult( full_key, timer0.diff_time(timer3)*1000.0 );
+            stat.AddRoundResult( proc_key, timer2.diff_time(timer3)*1e+3 );
+            stat.AddRoundResult( fill_key, timer1.diff_time(timer2)*1e+3 );
+            stat.AddRoundResult( init_key, timer0.diff_time(timer1)*1e+3 );
+            stat.AddRoundResult( full_key, timer0.diff_time(timer3)*1e+3 );
         }
         stat.SetStatisticFormula("1Average", "=AVERAGE(ROUNDS)");
         stat.SetStatisticFormula("2+/-", "=(MAX(ROUNDS)-MIN(ROUNDS))/2");
