@@ -1,29 +1,21 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2017 Intel Corporation
 
-    This file is part of Threading Building Blocks.
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
+
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 
-    Threading Building Blocks is free software; you can redistribute it
-    and/or modify it under the terms of the GNU General Public License
-    version 2 as published by the Free Software Foundation.
 
-    Threading Building Blocks is distributed in the hope that it will be
-    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with Threading Building Blocks; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    As a special exception, you may use this file as part of a free software
-    library without restriction.  Specifically, if other files instantiate
-    templates or use macros or inline functions from this file, or you compile
-    this file and link it with other files to produce an executable, this
-    file does not by itself cause the resulting executable to be covered by
-    the GNU General Public License.  This exception does not however
-    invalidate any other reasons why the executable file might be covered by
-    the GNU General Public License.
 */
 
 #if (_MSC_VER)
@@ -38,19 +30,9 @@
 #include "tbb_misc.h"
 #include "itt_notify.h"
 
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    // Suppress "C++ exception handler used, but unwind semantics are not enabled" warning in STL headers
-    #pragma warning (push)
-    #pragma warning (disable: 4530)
-#endif
-
 #include <cstring>
 #include <memory> //for uninitialized_fill_n
 
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    #pragma warning (pop)
-#endif
-
 #if defined(_MSC_VER) && defined(_Wp64)
     // Workaround for overzealous compiler warnings in /Wp64 mode
     #pragma warning (disable: 4267)
@@ -81,7 +63,7 @@ public:
 
     // TODO: optimize accesses to my_first_block
     //! assign first segment size. k - is index of last segment to be allocated, not a count of segments
-    inline static void assign_first_segment_if_neccessary(concurrent_vector_base_v3 &v, segment_index_t k) {
+    inline static void assign_first_segment_if_necessary(concurrent_vector_base_v3 &v, segment_index_t k) {
         if( !v.my_first_block ) {
             /* There was a suggestion to set first segment according to incompact_predicate:
             while( k && !helper::incompact_predicate(segment_size( k ) * element_size) )
@@ -138,8 +120,7 @@ public:
         } else {
             ITT_NOTIFY(sync_acquired, &s);
         }
-        if(s.load<relaxed>() != segment_allocated())
-            throw_exception(eid_bad_last_alloc); // throw custom exception, because it's hard to recover correctly after segment_allocation_failed state
+        enforce_segment_allocated(s.load<relaxed>()); //it's hard to recover correctly after segment_allocation_failed state
         return s;
     }
 
@@ -204,8 +185,7 @@ public:
         const void *arg;
         safe_init_body(internal_array_op2 init, const void *src) : func(init), arg(src) {}
         void operator()(segment_t &s, void *begin, size_type n) const {
-            if(s.load<relaxed>() != segment_allocated())
-                throw_exception(eid_bad_last_alloc); // throw custom exception
+            enforce_segment_allocated(s.load<relaxed>());
             func( begin, arg, n );
         }
     };
@@ -270,7 +250,7 @@ concurrent_vector_base_v3::size_type concurrent_vector_base_v3::helper::enable_s
     size_type size_of_enabled_segment =  segment_size(k);
     size_type size_to_allocate = size_of_enabled_segment;
     if( !k ) {
-        assign_first_segment_if_neccessary(v, default_initial_segments-1);
+        assign_first_segment_if_necessary(v, default_initial_segments-1);
         size_of_enabled_segment =  2 ;
         size_to_allocate = segment_size(v.my_first_block);
 
@@ -288,10 +268,11 @@ concurrent_vector_base_v3::size_type concurrent_vector_base_v3::helper::enable_s
             array0 = s[0].load<acquire>();
         }
         ITT_NOTIFY(sync_acquired, &s[0]);
-        if(array0 != segment_allocated()) { // check for segment_allocation_failed state of initial segment
-            publish_segment(s[k], segment_allocation_failed()); // and assign segment_allocation_failed state here
-            throw_exception(eid_bad_last_alloc); // throw custom exception
-        }
+
+        segment_scope_guard k_segment_guard(s[k], false);
+        enforce_segment_allocated(array0); // initial segment should be allocated
+        k_segment_guard.dismiss();
+
         publish_segment( s[k],
             static_cast<void*>(array0.pointer<char>() + segment_base(k)*element_size )
         );
@@ -367,7 +348,7 @@ void concurrent_vector_base_v3::internal_reserve( size_type n, size_type element
     if( n>max_size )
         throw_exception(eid_reservation_length_error);
     __TBB_ASSERT( n, NULL );
-    helper::assign_first_segment_if_neccessary(*this, segment_index_of(n-1));
+    helper::assign_first_segment_if_necessary(*this, segment_index_of(n-1));
     segment_index_t k = helper::find_segment_end(*this);
 
     for( ; segment_base(k)<n; ++k ) {
@@ -384,7 +365,7 @@ void concurrent_vector_base_v3::internal_copy( const concurrent_vector_base_v3&
     size_type n = src.my_early_size;
     __TBB_ASSERT( my_segment == my_storage, NULL);
     if( n ) {
-        helper::assign_first_segment_if_neccessary(*this, segment_index_of(n-1));
+        helper::assign_first_segment_if_necessary(*this, segment_index_of(n-1));
         size_type b;
         for( segment_index_t k=0; (b=segment_base(k))<n; ++k ) {
             if( (src.my_segment.load<acquire>() == src.my_storage && k >= pointers_per_short_table)
@@ -407,15 +388,14 @@ void concurrent_vector_base_v3::internal_assign( const concurrent_vector_base_v3
         size_type b=segment_base(k);
         size_type new_end = b>=n ? b : n;
         __TBB_ASSERT( my_early_size>new_end, NULL );
-        if( my_segment[k].load<relaxed>() != segment_allocated()) // check vector was broken before
-            throw_exception(eid_bad_last_alloc); // throw custom exception
+        enforce_segment_allocated(my_segment[k].load<relaxed>()); //if vector was broken before
         // destructors are supposed to not throw any exceptions
         destroy( my_segment[k].load<relaxed>().pointer<char>() + element_size*(new_end-b), my_early_size-new_end );
         my_early_size = new_end;
     }
     size_type dst_initialized_size = my_early_size;
     my_early_size = n;
-    helper::assign_first_segment_if_neccessary(*this, segment_index_of(n));
+    helper::assign_first_segment_if_necessary(*this, segment_index_of(n));
     size_type b;
     for( segment_index_t k=0; (b=segment_base(k))<n; ++k ) {
         if( (src.my_segment.load<acquire>() == src.my_storage && k >= pointers_per_short_table)
@@ -425,8 +405,8 @@ void concurrent_vector_base_v3::internal_assign( const concurrent_vector_base_v3
         helper::extend_table_if_necessary(*this, k, 0);
         if( my_segment[k].load<relaxed>() == segment_not_used())
             helper::enable_segment(*this, k, element_size);
-        else if( my_segment[k].load<relaxed>() != segment_allocated() )
-            throw_exception(eid_bad_last_alloc); // throw custom exception
+        else
+            enforce_segment_allocated(my_segment[k].load<relaxed>());
         size_type m = k? segment_size(k) : 2;
         if( m > n-b ) m = n-b;
         size_type a = 0;
@@ -483,8 +463,7 @@ concurrent_vector_base_v3::size_type concurrent_vector_base_v3::internal_grow_to
                 backoff.pause();
             ITT_NOTIFY(sync_acquired, &s);
         }
-        if( my_segment[i].load<relaxed>() != segment_allocated() )
-            throw_exception(eid_bad_last_alloc);
+        enforce_segment_allocated(my_segment[i].load<relaxed>());
     }
 #if TBB_USE_DEBUG
     size_type capacity = internal_capacity();
@@ -502,7 +481,7 @@ concurrent_vector_base_v3::size_type concurrent_vector_base_v3::internal_grow_by
 void concurrent_vector_base_v3::internal_grow( const size_type start, size_type finish, size_type element_size, internal_array_op2 init, const void *src ) {
     __TBB_ASSERT( start<finish, "start must be less than finish" );
     segment_index_t k_start = segment_index_of(start), k_end = segment_index_of(finish-1);
-    helper::assign_first_segment_if_neccessary(*this, k_end);
+    helper::assign_first_segment_if_necessary(*this, k_end);
     helper::extend_table_if_necessary(*this, k_end, start);
     helper range(my_segment, my_first_block, element_size, k_start, start, finish);
     for(; k_end > k_start && k_end >= range.first_block; --k_end ) // allocate segments in reverse order
